
import json
from dockerfile_parse import DockerfileParser

ModelParameterDefaults = {
    'mirostat': 0,
    'mirostat_eta': 0.1, 
    'mirostat_tau': 5.0,
    'num_ctx': 4096,
    'num_gqa': 1,
    'num_gpu': 1,   
    'num_thread': 8,
    'repeat_last_n': 64,
    'repeat_penalty': 1.0,
    'temperature': 0.7,
    'seed': 0,
    'stop': "AI assistant:",
    'tfs_z': 1,
    'num_predict': 128,
    'top_k': 40,
    'top_p': 0.9,
    'rope_frequency_base': 1e+6
}

class Modelfile:

    def __init__(self, 
                model_reference: str=None,
                modelfile_text: str=None, 
                template: str=None,
                messages: list=[],
                system_message: str=None, 
                parameters: dict={}, 
                **kwargs):
        
        self.messages = messages
        self.parameters = parameters
        self.model = model_reference
        self.model_reference = model_reference        
        self.template = template
        self.system_message = system_message

        if modelfile_text is not None:
            self._from_modelfile_text(modelfile_text)

        if kwargs is not None:
            self.set_parameters(kwargs) 
    
    def __eq__(self, value: object) -> bool:
        if not isinstance(value, Modelfile):
            return False
        
        return self.model_reference == value.model_reference

    def _wrap(self, value:str, single_quote: bool=False) -> str:
        if value is not None:            
            value = value.strip()
            if len(value) != 0:                    
                if not value.startswith('\"'):
                    if single_quote:
                        value = f"\"{value}\""
                    else:
                        value = f'\"\"\"{value}\"\"\"'

        return value
    
    def _from_modelfile_text(self, modelfile_text: str):
        dfp = DockerfileParser()
        dfp.content = modelfile_text
        ret = json.loads(dfp.json)
        for obj in ret:
            key = next(iter(obj)).upper()
            value = obj[key]
            if key == 'PARAMETER':
                # split the value
                toks = value.split(" ")
                param_name = toks[0]
                if len(toks) == 2:
                    param_value = Modelfile.cast_to_type(param_name, toks[1])
                else:
                    param_value = " ".join(toks[1:])
                # handle multiple stop parameters
                if param_name == 'stop':
                    if self.parameters.get(param_name) is not None:
                        self.parameters[param_name].append(param_value)
                    else: 
                        self.parameters[param_name] = [param_value]
                else:
                    # otherwise, only one param is allowed 
                    #   therefore the last key will set the final value
                    self.parameters[param_name] = param_value
            elif key == 'TEMPLATE':
                self.template = value
            elif key == "FROM":
                self.model = value
                if self.model_reference is None:
                    self.model_reference = self.model
            elif key == "SYSTEM":
                self.system_message = value
            elif key == "MESSAGE":
                self.messages.append(value)

    def to_modelfile(self, filename: str):
        with open(filename, 'w') as f:
            f.write(self.to_modelfile_text())

    @staticmethod
    def from_modelfile(filename: str) -> 'Modelfile':
        from os import path
        if path.exists(filename):
            with open(filename, 'r') as f:
                return Modelfile(modelfile_text=f.read())
            
        raise FileNotFoundError(f"File not found: {filename}")
    
    @staticmethod
    def cast_to_type(key, str_value):
        """Casts a string value to the type of the corresponding key in a dictionary.

        Args:
            key: The key in the dictionary.
            str_value: The value to be cast, as a string.

        Returns:
            The cast value.

        Raises:
            ValueError: If the key is not found in the dictionary or if the cast fails.
        """

        if key not in ModelParameterDefaults:
            raise ValueError(f"Key {key} not found in ModelParameterDefaults")

        try:
            return type(ModelParameterDefaults[key])(str_value)
        except (ValueError, TypeError):
            raise ValueError(f"Failed to cast value '{str_value}' for key {key}")
                        

    def to_modelfile_text(self):
        # start with Model name
        return_string = f"FROM {self.model}\n"

        # continue to parameters section
        return_string += self.parameter_string + '\n' 
    
        if self.template:
            return_string += f"TEMPLATE {self._wrap(self.template)}\n"

        if self.system_message:
            return_string += f"SYSTEM {self._wrap(self.system_message)}\n"

        if self.messages:
            return_string += '\n'.join([f"MESSAGE {self._wrap(message)}" for message in self.messages]) + '\n'

        return return_string
        
    @property
    def parameter_string(self):
        for key, value in self.parameters.items():
            arr = []
            if isinstance(value, list):
                arr.extend([f"PARAMETER {key} {self._wrap(v, single_quote=True)}" for v in value])
            else:
                arr.append(f"PARAMETER {key} {value}")

        return "\n".join(arr)

    def set_parameters(self, params: dict):
        for key, value in params.items(): 
              self.set_parameter(key, value)

    def set_parameter(self, key, value):
        if key not in ModelParameterDefaults:
            raise ValueError(f"Invalid parameter name: {key}")
        
        if key == 'stop' and isinstance(value, str):
            # use array
            value = [value]

        self.parameters[key] = value




import os
from ollama_hydra.modelfile import Modelfile

def test_stop_w_spaces():
    modelfile = Modelfile.from_modelfile('tests/modelfiles/t2.modelfile')
    assert modelfile.parameters['stop'] == ['\"[INST] Hello World\"','\"[/INST]\"']
    assert modelfile.model == 'mistral:latest'

def test_modelfile_init():
    modelfile = Modelfile.from_modelfile('tests/modelfiles/t1.modelfile')

    assert modelfile.model ==  'codellama:latest'
    assert modelfile.parameters['temperature'] == 0.5
    assert modelfile.system_message is not None

def test_serialization():
    modelfile = Modelfile.from_modelfile('tests/modelfiles/t1.modelfile')

    modelfile_2 = Modelfile(modelfile_text=modelfile.to_modelfile_text())

    assert modelfile.model == modelfile_2.model
    assert modelfile.system_message == modelfile_2.system_message
    assert modelfile.parameters['temperature'] == modelfile_2.parameters['temperature']
    assert modelfile.to_modelfile_text() == modelfile_2.to_modelfile_text()


    
def test_messages_section():
    modelfile = Modelfile(model="mistral:latest")    

    modelfile.messages.append("Message 1")
    modelfile.messages.append("Message 2")

    text = modelfile.to_modelfile_text()
    assert "Message 1" in text
    assert "Message 2" in text


def test_stop_parameter_init():
    modelfile = Modelfile(model="mistral:latest", stop="[INST]")
    assert modelfile.parameters['stop'] == ["[INST]"]



def test_multiple_stop():
    modelfile = Modelfile(model="mistral:latest", stop=["[INST]", "[/INST]"])
    assert modelfile.parameters['stop'] == ["[INST]", "[/INST]"]








import asyncio
import requests
import ollama
import json
from time import sleep
import os
from enum import Enum
import functools
from ollama_hydra.modelfile import Modelfile
from ollama_hydra.client import AutogenClient

model_port_mappings = {}  
ollama_port = os.environ.get("OLLAMA_PORT", 11434)  # Port for ollama.list() API
ollama_host = os.environ.get("OLLAMA_HOST", '127.0.0.1') 
next_available_port = os.environ.get("NEXT_AVAILABLE_PORT", 11435)  # Starting port for dynamic assignment


print(f"Connecting to OLLAMA on {ollama_host}")

        
ollama_url = f'http://localhost:{ollama_port}'

clients = [AutogenClient(ollama_host=ollama_host, ollama_port=ollama_port)]


class ModelMapping:

    def __init__(self, modelfile: Modelfile, port: int) -> None:        
        self.port = port
        self.modelfile = modelfile

    @staticmethod
    def create_mapping(model_reference: str, port: int) -> 'ModelMapping':
        if model_reference is not None:
            show_ret = ollama.show(model_reference)
            modelfile_text = show_ret['modelfile']
            mf = Modelfile(model_reference=model_reference, modelfile_text=modelfile_text)
        else:
            mf = None
        
        mm = ModelMapping(modelfile=mf, port=port)
        return mm
   
    def __eq__(self, value: object) -> bool:
        if not isinstance(value, ModelMapping):
            return False
        
        return self.port == value.port and \
            self.modelfile == value.modelfile
    
    @property
    def model(self):
        if self.modelfile is not None:
            return self.modelfile.model_reference
        
        return None



async def handle_request(reader, writer, port):
    try:
        request_line = await reader.readline()
        method, path, _ = request_line.decode().strip().split()
        function_name = path[1:]

        headers = {}
        while True:
            header_line = await reader.readline()
            if header_line == b'\r\n':
                break
            header_key, header_value = header_line.decode().strip().split(':', 1)
            headers[header_key] = header_value.strip()

        # Gather request data
        request_data = {}
        content_length = int(headers.get('Content-Length', 0))
        if content_length > 0:
            request_body = await reader.read(content_length)
            if headers.get('Content-Type', '').startswith('application/json'):
                request_data = json.loads(request_body.decode())
            else:
                request_data = request_body.decode()

        modelfile = model_port_mappings[port]
        
        selected_client = None
        for client in clients:
            if hasattr(client, function_name):
                selected_client = client
                break

        if selected_client:
            client_func = getattr(selected_client, function_name)
            model = modelfile.model          
            if model is None and "model" in request_data:
                model = request_data.get("model", None)
                            
            if model is not None:
                response_data = client_func(request_data, model=model)
            else:
                response_data = {
                    "message": "'model' is required parameter. Either use a mapped port, or if unmapped pass valid ollama 'model' reference",
                    "code": 404
                }
            if isinstance(response_data, dict) and "code" in response_data:
                status_code = response_data.get('code', 200)
                if status_code == 200:
                    reason = "OK"
                else:
                    reason = "Error"
            else:
                status_code = 200
                reason = "OK"
        else:        
            response_data = { "message": f"No client found to handle the request on `{path}`.", "code": 404 }
            status_code = 404
            reason = "Not Found"
    except Exception as e:
        error_message = f"Error handling request: {e}"
        response_data = { "message": error_message, "code": 503 }
        status_code = 503
        reason = "Server Error"            

    try:
        # Convert response data to JSON
        if hasattr(response_data, 'to_json'):
            response_data = response_data.to_json()
    except Exception as e:
        em = f"Error parsing object of type {type(response_data)}"
        print(em)
        response_data = {"message": em, "code": 503}
        status_code = 503
        reason = "Server Error"

    try:       
        response_body = json.dumps(response_data).encode()

        # Write response headers
        writer.write(f"HTTP/1.1 {status_code} {reason}\r\n".encode())
        writer.write(b"Content-Type: application/json\r\n")
        writer.write(f"Content-Length: {len(response_body)}\r\n".encode())
        writer.write(b"\r\n")

        # Write response body
        writer.write(response_body)
        await writer.drain()
        writer.close()

    except Exception as e:
        print(f"Error writing request to caller {e}")
        

def _update_model_mapping_helper():        
    global model_port_mappings, next_available_port
    try:
        models = ollama.list()
        models = models.get('models', [])
        new_mappings = {}
        port = next_available_port
        new_mappings[port] = ModelMapping.create_mapping(None, port)
        port += 1
        for model in models:            
            port += 1
            new_mappings[port] = ModelMapping.create_mapping(model['model'], port)

        if new_mappings != model_port_mappings:
            print("\n\nModel Mappings:\n")
            for p, m in new_mappings.items():
                if m.model is not None:
                    print(f"\t{m.port} -> {m.model}")
                else:
                    print(f"\t{m.port} -> OPEN // use 'model' to set on a call-by-call basis")

        model_port_mappings = new_mappings
        
    except Exception as e:
        print(f"Error updating model mappings: {e}")

async def start_servers(host, port):
    # Create a partial function with the port number pre-filled
    handle_request_with_port = functools.partial(handle_request, port=port)

    server = await asyncio.start_server(handle_request_with_port, host, port)
    await server.serve_forever()


def print_startup():
    print("\n\n")
    print(("-" * 50))
    print("Hydra Proxy Startup Complete")
    print(("-" * 50))
    print("\n\n")


def enable_sockets():
    try:
        host = '127.0.0.1'        
            
        if len(model_port_mappings) == 0:
            print("Unable to find any mappings. Exiting.")            
            # exit program
            return
        
        loop = asyncio.new_event_loop()
        for si in model_port_mappings.keys():
            i = int(si)
            loop.create_task(start_servers(host, i))

        print_startup()

        loop.run_forever()
    except Exception as exc:
        print(exc)


if __name__ == "__main__":
    _update_model_mapping_helper()
    enable_sockets()



from setuptools import setup, find_packages

setup(
    name='ollama-hydra',
    version='0.1.7',
    description='Proxy server for managing multiple Ollama models for AutoGen Agents', 
    long_description='', 
    author='Cyrus Radfar',
    author_email='support@v1.co',  
    url='https://github.com/sendthis-ai/ollama-hydra',
    packages=find_packages(),
    install_requires=[
        'requests',
        'ollama', 
        'pyautogen',
        'dockerfile-parse'
    ],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: OS Independent",
    ],
    entry_points={
        'console_scripts': [
            'ollama-hydra = ollama_hydra.proxy:main',
            'ohydra = ollama_hydra.proxy:main'  # Added alias
        ],
    },
)



import ollama
from typing import List, Optional, Union, Dict, Any
from dataclasses import dataclass, field, asdict
from autogen.oai.client import ModelClient
import json

class OllamaClient(ModelClient):
    
    def __init__(self, ollama_host:str = '127.0.0.1', 
                 ollama_port: int = 11434, 
                 **kwargs) -> None:
        super().__init__(kwargs)
        self.ollama_host = ollama_host
        self.ollama_port = ollama_port


@dataclass
class OllamaMessage:
    content: Optional[str] = None
    function_call: Optional[dict] = None


@dataclass
class OllamaChoice:
    message: OllamaMessage

@dataclass
class ModelClientResponse:
    class Choice:
        @dataclass
        class Message:
            content: Optional[str] = None  # Initialize with None

        message: Message = field(default_factory=Message)

    choices: List[Choice] = field(default_factory=list)
    model: str = ""

    def to_json(self):
        return json.dumps(asdict(self))


class AutogenClient(OllamaClient):
    def __init__(self,**kwargs):
        super().__init__(kwargs)
        

    def create(self, params: Dict[str, Any], model) -> ModelClient.ModelClientResponseProtocol:
         
        num_of_responses = params.get("n", 1)
        messages = params.get("messages", [])        
        system_message = params.get("system", None)
        stream = params.get("stream", False)
        prompt = params.get("prompt", None)

        # error handling // gating
        if prompt is not None and messages:
            return {
                "message": "Error, use either 'messages' or 'prompt' but not both",
                "code": 404
            }
        elif prompt is None and not messages:
            return {
                "message": "Must pass 'messages' or 'prompt'",
                "code": 404
            }        

        response = ModelClientResponse()
        response.choices = []
        response.model = model

        args = {
            "model": model            
        }
        if prompt is not None:
            args['prompt'] = prompt
            if system_message is not None:
                args['system'] = system_message
        else:
            args['messages'] = messages
            if system_message is not None:
                return {
                    "message": "'system' isn't valid when using 'messages', only when generating with 'prompt'",
                    "code": 404
                }        

        if messages:
            # Use ollama.chat for conversation-style interactions
            chat_response = ollama.chat(**args)
            for _ in range(num_of_responses):
                choice = OllamaChoice(message=OllamaMessage(content=chat_response['message']['content']))
                response.choices.append(choice)
        else:
            # Use ollama.generate for single-prompt generations            
            for _ in range(num_of_responses):
                output = ollama.generate(**args)  #), max_length=max_length)
                choice = OllamaChoice(message=OllamaMessage(content=output))
                response.choices.append(choice)

        return response

    def message_retrieval(
        self, response: ModelClient.ModelClientResponseProtocol
    ) -> Union[List[str], List[ModelClient.ModelClientResponseProtocol.Choice.Message]]:
        choices = response.choices
        return [choice.message for choice in choices]

    def cost(self, response: ModelClient.ModelClientResponseProtocol) -> float:
        # Assuming Ollama doesn't provide cost information, return 0
        return 0.0

    @staticmethod
    def get_usage(response: ModelClient.ModelClientResponseProtocol) -> Dict:
        # Assuming Ollama doesn't provide usage information, return an empty dict
        return {}



